/**
 *  \file			comm.h
 *  \author 		Tranchida Giulio, No Matricola 241732\n
 *  \brief 			libreria di comunicazione server / client
 *
 *  Libreria che definisce l'interfaccia di comunicazione fra client e server
 *  (canali di comunicazione realizzati con socket AF_UNIX)
 *
 *  This program is free software; you can redistribuite it and/or modify
 *  it under the terms of the GNU/General Pubblic License as published the
 *  Free software Foundation; either version 2 of the License, or
 *  (at your opinion) any later version.
 * */
#ifndef COMM_H
#define COMM_H

#include <sys/types.h>
#include <sys/socket.h>	/* definitions of structures needed for sockets */
#include <netinet/in.h>	/* constants and structures needed for internet domain addresses */
#include <sys/poll.h> 	/* poll(2) and related definitions. */
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>

#include "msg.h"
#include "errore.h"

/* ----= TIPI =---- */

/** Struttura delle sockets di ascolto del server */
typedef struct {
	size_t		  skDim;	/**< Numero di socket in ascolto */
	struct pollfd *sk;		/**< Data structure describing a polling request*/
} comm;

#define MAXCONN 20			/**< Numero di connessioni massime accettate */
#define INVALID_DESC -1		/**< Invalid file descriptor (server socket) */

/* -= FUNZIONI =- */

/** Crea le sockets di ascolto.
 *  Open passive (server) sockets for the indicated inet service & protocol.
 *  Notice in the last sentence that "sockets" is plural.  During the interim
 *  transition period while everyone is switching over to IPv6, the server
 *  application has to open two sockets on which to listen for connections...
 *  one for IPv4 traffic and one for IPv6 traffic.
 *
 *  \param service   Pointer to a character string representing the
 *  				 well-known port on which to listen (can be a service
 *  				 name or a decimal number).
 *  \return
 *     - a pointer to a new comm structure, on succes
 *     - a NULL pointer, on failure and sets errno */
comm * newServerChannel(const char *service);

/** Accetta una connessione da parte di un client
 *  \param srvComm Struttura delle sockets di ascolto del server
 *  \return
 *       - il descrittore del canale di trasmissione con il client
 *       - -1 in casi di errore (sets errno)
 */
int acceptConnection(comm * srvComm);

/** Distrugge le sockets di ascolto del server
 *  \param srvComm Struttura delle sockets del server da distruggere
 *  \return
 *         -  0  se tutto ok, 
 *         - -1  se errore (sets errno)
 */
int closeServerSocket(comm * srvComm);

/** Legge un messaggio dal canale di trasmissione
 *  \param sc  channel di trasmissione
 *  \param msg  struttura che conterra' il messagio letto 
 *		(deve essere allocata all'esterno della funzione,
 *		tranne il campo buffer)
 *  \return
 *     -  -1    se errore (errno settato),
 *     -  SEOF  se EOF sul socket 
 *     -  lung  lunghezza del buffer letto, se OK
 */
int receiveMessage(int sc, message_t *msg);

/** Scrive un messaggio sul canale di trasmissione
 *  \param  sc channel di trasmissione
 *  \param msg struttura che contiene il messaggio da scrivere
 *  \return
 *     -  -1   se errore (sets errno) 
 *     -  n    il numero di caratteri inviati (se OK)
 */
int sendMessage(int sc, message_t *msg);

/** Chiude la socket del client
 *  \param  sc il descrittore del channel da chiudere
 *  \return
 *       - 0  se tutto ok
 *       - -1 se errore (sets errno)
*/
int closeConnection(int sc);

/** Crea una socket di trasmissione verso il server, tentando la
 *  connessione per un massimo di due minuti se MAXSLEEP e' 128
 *  \param  hostname Indirizzo del server
 *  \param  service  Nome del servizio (noto) o numero di porta del server
 *  \return
 *      -  c (c>0) la socket di comunicazione del client
 *      - -1 in casi di errore (sets errno) */
int retryConnection(const char* hostname, const char * service);

#endif
