/**
 *  \file			commSSL.h
 *  \author 		Tranchida Giulio, No Matricola 241732\n
 *  \brief 			libreria di comunicazione server client\n\n
 *
 *  Libreria che definisce l'interfaccia di comunicazione fra client e server
 *  (canali di comunicazione realizzati con socket AF_UNIX)
 *
 * This program is free software; you can redistribuite it and/or modify
 * it under the terms of the GNU/General Pubblic License as published the
 * Free software Foundation; either version 2 of the License, or
 * (at your opinion) any later version.
 * */

#ifndef COMMSSL_H_
#define COMMSSL_H_

#include "comm.h"

#include <openssl/rand.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

/** Certificato RSA del server */
#define RSA_SERVER_CERT     "server.crt"
/** Chiave del server */
#define RSA_SERVER_KEY      "server.key"
/** Certificato RSA della Certification autority del server */
#define RSA_SERVER_CA_CERT  "server_ca.crt"
/** Percorso del certificato della CA del server */
#define RSA_SERVER_CA_PATH  "sys$common:[syshlp.examples.ssl]"

/** Certificato RSA del client */
#define RSA_CLIENT_CERT		"client.crt"
/** Chiave del client */
#define RSA_CLIENT_KEY		"client.key"
/** Certificato RSA della Certification autority del client */
#define RSA_CLIENT_CA_CERT	"client_ca.crt"
/** Percorso del certificato della CA del client */
#define RSA_CLIENT_CA_PATH	"sys$common:[syshlp.examples.ssl]"

#define ON   1
#define OFF  0

/* ----= TIPI =---- */

/** Struttura dati di comunicazione del server */
typedef struct {
	comm * srvSk;		/**< Canale di ascolto del server (IPv4/IPv6) */
    SSL_CTX *ctx;		/**< Contesto SSL per il Server */
} srvSkSSL;

/** Struttura di comunicazione per la connessione del client */
typedef struct {
    int socket;			/**< Canale di comunicazione del client */
    SSL_CTX *ctx;		/**< Contesto SSL per il client */
    SSL *ssl;			/**< Handler SSL associato al canale del client */
} cli_conn;


/* ----= FUNZIONI =---- */

/** Crea le sockets di ascolto.
 *  Open passive (server) sockets for the indicated inet service & protocol.
 *  Notice in the last sentence that "sockets" is plural.  During the interim
 *  transition period while everyone is switching over to IPv6, the server
 *  application has to open two sockets on which to listen for connections...
 *  one for IPv4 traffic and one for IPv6 traffic.
 *
 *  \param service   Pointer to a character string representing the
 *  				 well-known port on which to listen (can be a service
 *  				 name or a decimal number).
 *  \return
 *     - a pointer to a new comm structure, on succes
 *     - a NULL pointer, on failure and sets errno */
srvSkSSL * newSSLServerChannel(const char *service);

/** accetta una connessione da parte di un client
 *  \param server struttura dati della socket di ascolto del server
 *  \return
 *     - a pointer to a new 'cli_conn' structure, on succes
 *     - a NULL pointer, on failure and sets errno */
cli_conn * acceptSSLConnection(srvSkSSL * server);

/** legge un messaggio dal channel di trasmissione
 *  \param client struttura di comunicazione del cliente
 *  \param msg  struttura che conterra' il messagio letto
 *		(deve essere allocata all'esterno della funzione,
 *		tranne il campo buffer)
 *  \return
 *     -  -1    se errore (errno settato),
 *     -  SEOF  se EOF sul socket
 *     -  lung  lunghezza del buffer letto, se OK
 */
int receiveSSLMessage(cli_conn * client, message_t * msg);

/** scrive un messaggio sul channel di trasmissione
 *   \param client struttura di comunicazione del client
 *   \param msg struttura che contiene il messaggio da scrivere
 *   \return
 *     -  -1   se errore (sets errno)
 *     -  n    il numero di caratteri inviati (se OK)
 */
int sendSSLMessage(cli_conn * client, message_t * msg);

/** Distrugge un canale di ascolto (server-side)
 *   \param server struttura dati della socket di ascolto del server
 *   \return
 *     -  0  on succes,
 *     - -1  se errore (sets errno)
 */
int closeSSLServerSocket(srvSkSSL * server);

/** Chiude un socket
 *  \param  client struttura di comunicazione del client
 *  \return
 *     - 0  on succes,
 *     - -1 se errore (sets errno)
*/
int closeSSLConnection(cli_conn * client);

/** crea una socket di trasmissione verso il server, tentando la
 *  connessione per un massimo di due minuti se MAXSLEEP e' 128
 *  \param  hostname Indirizzo del server
 *  \param  service	 Nome del servizio (noto) o numero di porta del server
 *  \return
 *     - a pointer to a new 'cli_conn' structure, on succes
 *     - a NULL pointer, on failure and sets errno */
cli_conn * retrySSLConnection(const char * hostname, const char * service);

/** Cerca di scaricare tutta la merda messa in memoria dalle librerie SSL */
void unloadSSL();

#endif /* COMMSSL_H_ */
