/**
 * 	\file		pool.h
 *	\author		Tranchida Giulio, No Matricola 241732\n
 *				Si dichiara che il contenuto di questo file e',
 * 				in ogni sua parte, opera originale dell'autore.\n\n
 *
 * This program is free software; you can redistribuite it and/or modify
 * it under the terms of the GNU/General Pubblic License as published the
 * Free software Foundation; either version 2 of the License, or
 * (at your opinion) any later version.
 * */
#ifndef POOL_H_
#define POOL_H_

#include <stdlib.h>
#include <pthread.h>
#include <signal.h>

/** Lista dei job (a solo uso interno) */
typedef struct joblist{
	/** Funzione eseguita dal thread worker */
	void *(*start_routine)(void*);
	/** Parametri della funzione */
	void * arg;
	/** puntatore all'elemento successivo della lista */
	struct joblist * next;
} joblist;

/** Struttura contentente le info per ogni pool di threads */
typedef struct ThreadPool{
	/** Variabile per porre il pool in fase di terminazione */
	volatile sig_atomic_t quitflag;
	/** Variabile per uccidere i threads */
	volatile sig_atomic_t shutdown;
	sigset_t * signalmask;		/**< Maschera dei segnali */
	int pool_size;				/**< Dimensione del pool di threads */
	int job_max;				/**< Dimensione massima della coda dei jobs */
	int job_size;				/**< Dimensione della coda dei jobs */
	joblist * firstjob;			/**< testa della coda dei jobs */
	joblist * lastjob;			/**< ultimo elemento della coda dei job */
	pthread_t * tids;			/**< Array dei tids dei threads */
	/** Mutex per l'accesso in mutua esclusione sulla coda dei jobs */
	pthread_mutex_t lock;
	pthread_cond_t notempty;	/**< Condizione "coda dei job non vuota" */
	pthread_cond_t empty;		/**< condizione "coda dei job vuota" */
} th_pool;

/** <H3>poolInit</H3>
 * Funzione che inizializza e ritorna un Pool di Threads\n
 *
 * \param dim Dimensione del pool di threads
 * \param job_max Dimensione massima della coda dei jobs
 * \param *set Opzionale. Maschera dei segnali che il pool usera'
 * 		  durante la creazione dei threads. Se NULL viene ignorata.
 * @returns
 * 		th_pool * Una nuovo pool di threads\n
 * 		NULL e setta errno se dim < 1
 * */
th_pool * poolInit(int dim, int job_max, sigset_t *set);

/**
 * <H3>poolDispatcher</H3>
 * Funzione che sottopone un nuovo job al pool di threads\n
 *
 * @param *pool Struttura dati del pool di thread
 * @param *(*start_routine)(void*) Il job da fare eseguire al pool
 * @param *arg argomento passato al job
 * @returns
 * 		pthread_t Il thread id a cui e' stato assegnato il job\n
 *		-1 se *pool e' NULL e setta errno\n
 *		-2 se  *start_routine e' NULL e setta errno\n
 *		-3 se fallisce l'allocazione di un nuovo job\n
 *		-4 se il pool di thread e' in fase di terminazione (rifiuta il job).
 *		-5 se il numero di job attualmente pendenti nella coda e' uguale
 *		   al numero massimo possibile di job in attesa.
 * */
int poolDispatcher(th_pool *pool, void *(*start_routine)(void*), void *arg);

/**
 * <H3>poolDestroy</H3>
 * Distrugge un Pool di Threads: uccide i threads e libera le risorse\n
 * In particolare:
 * - blocca la possibilita' di sottomettere un nuovo job al pool,\n
 * - aspetta che tutti i job pendenti vengano eseguiti,\n
 * - esegue la join di tutti i threads,\n
 * - libera le risorse e esce.\n</dd>
 *
 * @param *pool Il pool di threads da distruggere
 * @returns
 *		0 se tutto e' andato bene\n
 *		-1 se *pool e' NULL e setta errno
 * */
int poolDestroy(th_pool *pool);

#endif /* POOL_H_ */
